#!/usr/local/bin/perl
# $Id$
# Originally written by Jason Rennie <jr6b@andrew.cmu.edu>
# some modifications by Chris Browne <cbbrone@hex.net>

# If you have a large number of messages, or do not have a lot of physical
# memory for your machine, search for "DB_File" and uncomment the lines which
# will attach certain associative arrays to disk.

# This script will search through all your current EXMH/MH e-mail messages
# and create an .idata accordingly.  This way, instead of needing to refile
# a number of messages before ifile knows your preferences, ifile will have
# an instant knowledge base of your interests

# NOTE: This script will destroy your current .idata file (if one exists).
#       A backup will be created (kept?) as ".idata~".
#       If you do not keep extensive amounts of old mail, this will create a
#       more current version of your preferences.  If you keep old mail, the
#       new .idata file will include all the information in the old .idata
#       file (and then some).

# Any directory which contains a .skip_me file will be excluded from the
#   knowledge base.  Your drafts folder (as defined in your .mh_profile)
#   will be automatically skipped.

# This can be a very cpu/memory intensive script, especially if you have more
#   than just a few hundred messages.

# Note: cbbrowne@hex.net finds that his 6500 messages took 40 minutes on a
# 6x86-150+, 32MB RAM.  He thinks that a whole lot of this time is being
# consumed by the "aging" procedure.

# Running of this script will likely result in a larger ~/.idata file.
#   However, the size should remain fairly stable after this script is run.

# Changes June 26/97 - Christopher Browne - cbbrowne@hex.net
# A) Added in a locking scheme.

# It's not perfect (possible race condition between the time that we check 
# whether or not .idata.write.lock is already locked and the moment that we 
# actually create it...) but it's a tremendous improvement over the present
# situation.  Even if irefile does not cooperate by checking for .idata.lock
# (first, it should check for it, wait a couple of seconds and try again if
# it's not found, and then abort), we still cut vulnerability to a tiny
# fraction of what it used to be...

# B) Keeping the "operational" .idata sacrosanct while a new one is built
# The *far* more critical change is that the program *never* writes to a
# file called .idata.  It writes to .idata.$$, and once *all* processing is
# complete, it renames the old copy of .idata to .idata.old, and renames
# .idata.$$ to .idata.  Even ignoring any attempts to lock files, this change 
# alone cuts the "window of vulnerability" (the time during which .idata is 
# not in a usable state) down from as much as 40 minutes (whilst this script 
# is running) to the split second it takes to rename a file.

# C) "Aging" is only done once, shortly before the .idata file is written.
# This vastly improves performance beyond what it was previously.

# D) Tightened up the "line processing loop" (e.g. - where lines get split
# into words, and then inserted into the dictionary).  Hopefully a net
# improvement on the *other* major chunk of CPU crunching...

open(TMP, ">> /tmp/knowledge_base.info")
    || warn "Could not open /tmp/knowledge_base.info: $!\n";

print TMP "knowledge_base.perl called\n";

$home_dir = $ENV{"HOME"}."/";

$nuked = 0; # How many words were "nuked"
print TMP "Home dir = $home_dir\n";

# Define locks...
$w_lock = $home_dir . ".idata.write.lock";
$rw_lock = $home_dir . ".idata.lock";

# If lock file exists, see if the PID exists.  If so, abort.  Some other
# writer is in progress, and we shouldn't have a second one running.

# Check "total lock" first
&die_if_rw_lock_is_set($rw_lock);
&die_if_other_writer_running($w_lock);

# This should be improved on...  There's a possible race condition here,
# although it is mitigated by the fact that the program checks for the

# $rw_lock again immediately before renaming .idata.$$ to .idata.  The only
# danger that should remain here if two "writers" ran simultaneously is that
# we would lose the effects on .idata from one of the two.

open(WLOCK, ">$w_lock");
print WLOCK $$;  # Write Process ID out to lock file
close(WLOCK);

open(MH, "<$home_dir.mh_profile") 
    || &ouch("No .mh_profile found: $!\n");

while (<MH>) {
    if (m/Path:\s*(\S+)/) {
	$mail_path = $1;
	$mail_path .= "/";
    }
    if (m/draft-folder:\s*(\w+)/) {
	$draft_folder = $1;
    }
}

print TMP "Mail Path = $mail_path\n";

# Don't bother saving .idata; we're not going to touch it for quite a while
# yet...

#print TMP "cp $home_dir.idata $home_dir.idata~\n";
#system "cp $home_dir.idata $home_dir.idata~";

$tv[0] = 0;
$word[0] = "";

# Gets a listing of all the folders which contain messages

if ($mail_path =~ /^\//)
{
    $folders_file = $mail_path.".folders"; 
}
else
{ 
    $folders_file = $home_dir.$mail_path.".folders"; 
}

print "Opening $folders_file\n";

open(FOLDERS, "< $folders_file")
    || &ouch("Couldn't find .folders\n");

# Uncomment the following lines if you have a large number of messages 
# stored away or little physical memory on your machine.  This will help to
# conserve memory usage.

# use DB_File;
# use Fcntl;
# `rm /tmp/sum_nk /tmp/nk`;
# $dbtxn1 = tie %sum_nk, DB_File, "/tmp/sum_nk", O_RDWR|O_CREAT, 0640, $DB_BTREE;
# $dbtxn2 = tie %nk, DB_File, "/tmp/nk", O_RDWR|O_CREAT, 0640, $DB_BTREE;

while ($folder = <FOLDERS>)
{
    chop $folder;
    $tv[@tv] = $folder;
    $dest_folder = $folder;

    if ($mail_path =~ /^\//)
    { $mailbox_path = $mail_path.$folder; }
    else
    { $mailbox_path = $home_dir.$mail_path.$folder; }
    
    print TMP "Reading $mailbox_path/\n";
    print "Reading $mailbox_path/\n";
    
    if (opendir(BOX, "$mailbox_path/")
	&& ! -f "$mailbox_path/.skip_me"
	&& $folder ne $draft_folder)
    {
	while ($message = readdir(BOX))
	{
	    # checks to make sure it isn't dot-file or deleted file
	    # also doesn't consider files > 30,000 bytes (likely to contain
	    #   encoded attachements)
	    if ($message =~ m/^[^,\.]/ && $message =~ m/[^~]$/ 
		&& -f "$mailbox_path/$message"
		&& -s "$mailbox_path/$message" < 30000)
	    {
	        process_message($mailbox_path, $message);
		$docs{$dest_folder}++;
		$epoch++;

		$messages_handled++;
		if ($messages_handled % 200 == 0) {
		    print "Messages handled: $messages_handled  Words: $#word\n";
		}

		# only age words after a certain period of documents
		# $epoch keeps track of num. of documents since last aging
                &age_words;  
	    }
	}
	print TMP "\n";
	closedir(BOX);
    } else {
	if (-f "$mailbox_path/.skip_me" ||
	    $draft_folder eq $folder) {
	    print TMP "Skipping $mailbox_path/\n";
	    print "Skipping $mailbox_path/\n";
	} else {
	    print TMP "Did not open $mailbox_path/: $!\n";
	    print "Did not open $mailbox_path/: $!\n";
	}
    }
}

# One last time...
&age_words("FINISHED");

# Now, for some funky changes... 
# Put output into ~/.idata.$$ rather than into ~/.idata

# This has the salutory effect that ~/.idata stays in a usable form at all
# times, except for a brief instant during which it doesn't exist at all.

open(DATA, "> $home_dir.idata.$$")
    || &ouch("Couldn't open data file\n");

# prints folder names
for ($i=1; $i <= $#tv; $i++) 
{
    print DATA $tv[$i]." ";
}
print DATA "\n";

# prints total word frequency values
for ($i=1; $i <= $#tv; $i++) 
{
    if ($n{$tv[$i]} > 0) {
	print DATA $n{$tv[$i]}." ";
    } else {
	print DATA "0 ";
    }
}
print DATA "\n";

# prints number of documents for each category
for ($i=1; $i <= $#tv; $i++) 
{
    if ($n{$tv[$i]} > 0) {
	print DATA $docs{$tv[$i]}." ";
    } else {
	print DATA "0 ";
    }
}
print DATA "\n";

for ($i=1; $i <= $#word; $i++) 
   {
       print DATA $word[$i]." ".$age{$word[$i]}." ";
       
       for ($j=1; $j <= $#tv; $j++)
       {
	   if ($nk{$word[$i].",".$tv[$j]} > 0)
	   {
	       print DATA $j.":".$nk{$word[$i].",".$tv[$j]}." ";
	   }
       }
       print DATA "\n";
   }

close(DATA);

# Now, create a "rw" lock file (e.g. - lock .idata against *all* access),
# rename files, and then remove all locks.

# Abort if something has already turned the lock on...
&die_if_rw_lock_is_set($rw_lock);
# Create the second lock
open(RWLOCK, ">$rw_lock");
print RWLOCK $$;  # Process ID
close(RWLOCK);

# Now, start renaming files
# a) Preserve the old .idata file
rename "$home_dir.idata", "$home_dir.idata~";
# b) Rename new .idata file to .idata
rename "$home_dir.idata.$$", "$home_dir.idata";

# Clean up lock files
unlink $w_lock, $rw_lock;
################

print TMP "Rebuild of .idata complete.\n$nuked words were aged out of the dictionary.\n";
print "Rebuild of .idata complete.\n$nuked words were aged out of the dictionary.\n";

sub ouch 
{
    local($string) = @_;
    print TMP $string."\n";
    die "$string.: $!\n";
}

# If the PID *doesn't* exist, then something has died.  It *should* be safe
# to proceed, unless the 'total lock' ($rw_lock) is turned on.  In that 
# case, something very bad has happened, and we should warn the user, and just
# plain abort.
sub die_if_rw_lock_is_set 
{
    my($rw_lock) = @_;
    if (-e $rw_lock) {
	$oldpid = `cat $rw_lock`;
	chop $oldpid;
	open(PROCS, "ls /proc | grep $oldpid|");
	while (<PROCS>) {
	    if (/^0*$oldpid\s*/) {
# Another writer is running...
		&ouch("Another writer running as PID $oldpid\n");
      }
	}
	# Other process not found - BAD problem.
	&ouch("Another writer was running as PID $oldpid, but seems to have died.\n\nMake sure $home_dir.idata looks OK, remove $rw_lock, and try again.\n");
    }
    
}

sub die_if_other_writer_running 
{
    my($w_lock) =@_;
    if (-e $w_lock) 
    {
	$oldpid = `cat $w_lock`;
	chop $oldpid;
	open(PROCS, "ls /proc | grep $oldpid|");
	while (<PROCS>) {
	    if (/^0*$oldpid\s*/) {
		# Another writer is running...
		&ouch("Another writer already running as PID $oldpid\n");
	    }
	}
# Other process not found - minor problem.
	print TMP "Another writer was running as PID $oldpid, but seems to have died.\n";
	print TMP "Will overwrite $w_lock file.\n";
	print "Another writer was running as PID $oldpid, but seems to have died.\n";
	print "Will overwrite $w_lock file.\n";
    }
}

sub process_message {
    my($mailbox_path, $message) = @_;
    open(MSG, "$mailbox_path/$message")
	|| print TMP "Unable to open $mailbox_path/$message: $!\n";
   
    print TMP "$message ";
    
    local($line);
    
    # This section revisited - cbbrowne@hex.net
    # Some loops have been replaced by single statements, 
    # and substitutions have mostly been changed to operate
    # on the *whole* string so that they don't have to be
    # repeated 20-odd times per line.  This *ought* to improve
    # speed.  And this is the innermost loop here...
   while ($line = <MSG>)
   {
       my($prefix, $wrd);
       if ($line =~ s/subject://oi) { $prefix = "subject~"; }
       elsif ($line =~ s/from://oi) { $prefix = "from~"; }
       else {$prefix = ""};
       # Nuke out non-character stuff, turn it into single spaces...
       $line =~ tr/A-Za-z\.\@+/ /cs;  # now includes '@' and '.' characters
       $line = lc $line;          # Convert to lower case
       foreach $wrd (split(/ /, $line))
	   # By converting the non-char stuff into spaces, there is
	   # no longer a need to gradually split chunks off.  We can
	   # do it here in one fell swoop, saving us a heap of regex
	   # operations.  They're not overly inefficient, but they're
	   # certainly slower than the "no operations" alternative.
       {
	   # search for extra periods at the end of a word and remove them
	   $wrd =~ s/([^\.])\.*$/\1/o;
	   
	   # get rid of tiny words
	   next if (length($wrd) < 2); 
	   
	   $wrd = $prefix . $wrd;
	   if ($index{$wrd} == 0) 
	   {
	       $word[@word] = $wrd;
	       $index{$wrd} = $#word;
	   }
	   $nk{$wrd.",".$dest_folder}++;
	   $sum_nk{$wrd}++;
	   $n{$dest_folder}++;
       }
   }
    close(MSG);
    
}

sub age_words
{
    my($finished) = @_;

    # The important thing here is that we *don't* want to run thru and age
    # *all* the words *each* time we read a message.

    # One might consider only aging at the very end of the process of reading
    # messages, however, this will cause the problem of using an excessive
    # amount of memory.

    # With the current system, it will happen that words which are more or less
    # unique to a folder may have an unusually small age due to the sequence
    # in which messages are read and recorded
    

    if ($finished ne "FINISHED")
    {
	# change the following parameter if you'd like
	# 200 seems like a pretty good number
	$desired_aging = 200;
	if ($epoch++ < $desired_aging)
	{
	    return;
	}
    }
    else
    {
	$desired_aging = $epoch;
    }

    $epoch = 0;
    print "Aging...";
    for ($i=1; $i <= $#word; $i++) 
    {
	my($wrd) = $word[$i];
	$age{$wrd} += $desired_aging;
	
	# minimum word frequency required for word not to be nuked
	$cutoff_freq = (log($age{$wrd})/log(2)) - 1;
	
	# if the word is infrequent, remove it
	if ($sum_nk{$wrd} <= $cutoff_freq)
	{
	    $nuked++;
	    foreach $target (@tv)
	    {
		delete $nk{$wrd.",".$target};
	    }
	    delete $age{$wrd};
	    delete $index{$wrd};
	    delete $sum_nk{$wrd};
	    
	    $word[$i] = $word[$#word];
	    $index{$word[$i]} = $i;
	    $#word--;
	    $i--;
	}
    }
    print " nuked $nuked words (some words may have been counted more than once)\n";
}
